/*
 * gsttividenc1.c
 *
 * This file defines the "TIVidenc1" element, which encodes an xDM 1.x video
 * stream.
 *
 * Example usage:
 *     gst-launch videotestsrc num-buffers=10 !
 *         TIVidenc1 engineName="<engine name>" codecName="<codecName>" !
 *         fakesink silent=TRUE
 *
 * Notes:
 *  * This element currently assumes that video input is in  UYVY, Y8C8
 *    and NV12 formats.
 *
 *    Note: On DM6467, VDCE is used to perform Ccv from YUV422P semi to
 *    YUV420P semi.
 *
 * Original Author:
 *     Brijesh Singh, Texas Instruments, Inc.
 *
 * Copyright (C) 2008-2010 Texas Instruments Incorporated - http://www.ti.com/
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation version 2.1 of the License.
 *
 * This program is distributed #as is# WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <gst/gst.h>

#include <ti/sdo/dmai/Dmai.h>
#include <ti/sdo/dmai/VideoStd.h>
#include <ti/sdo/dmai/Buffer.h>
#include <ti/sdo/dmai/BufferGfx.h>
#include <ti/sdo/dmai/BufTab.h>
#include <ti/sdo/dmai/ce/Venc1.h>
#include <ti/sdo/dmai/Cpu.h>
#include <ti/sdo/dmai/Ccv.h>
#include <ti/sdo/dmai/ColorSpace.h>

#include "gsttividenc1.h"
#include "gsttidmaibuffertransport.h"
#include "gstticodecs.h"
#include "gsttithreadprops.h"
#include "gstticommonutils.h"
#include "gsttiquicktime_h264.h"

/* Declare variable used to categorize GST_LOG output */
GST_DEBUG_CATEGORY_STATIC (gst_tividenc1_debug);
#define GST_CAT_DEFAULT gst_tividenc1_debug

#define DEFAULT_BIT_RATE 2000000
#define INVALID_DEVICE   Cpu_Device_COUNT

/* Element property identifiers */
enum
{
  PROP_0,
  PROP_ENGINE_NAME,     /* engineName     (string)  */
  PROP_CODEC_NAME,      /* codecName      (string)  */
  PROP_RESOLUTION,      /* resolution     (string)  */
  PROP_BITRATE,         /* bitrate        (int)     */
  PROP_IN_COLORSPACE,   /* iColorSpace    (string)  */
  PROP_NUM_INPUT_BUFS,  /* numInputBuf    (int)     */
  PROP_NUM_OUTPUT_BUFS, /* numOutputBuf   (int)     */
  PROP_FRAMERATE,       /* frameRate      (int)     */
  PROP_DISPLAY_BUFFER,  /* displayBuffer  (boolean) */
  PROP_CONTIG_INPUT_BUF,/* contiguousInputFrame  (boolean) */
  PROP_GEN_TIMESTAMPS,  /* genTimeStamps  (boolean) */
  PROP_RATE_CTRL_PRESET,/* rateControlPreset  (gint) */
  PROP_ENCODING_PRESET, /* encodingPreset  (gint) */
  PROP_BYTE_STREAM      /* byteStream      (gboolean) */

};

/* Define source (output) pad capabilities.  Currently, MPEG2/4 and H264 are 
 * supported.
 */
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS
    ("video/mpeg, " 
     "mpegversion=(int){ 2, 4 }, "  /* MPEG versions 2 and 4 */
         "systemstream=(boolean)false, "
         "framerate=(fraction)[ 0, MAX ], "
         "width=(int)[ 1, MAX ], "
         "height=(int)[ 1, MAX ] ;"
     "video/x-h264, "                             /* H264                  */
         "framerate=(fraction)[ 0, MAX ], "
         "width=(int)[ 1, MAX ], "
         "height=(int)[ 1, MAX ]"
    )
);

/* Define sink (input) pad capabilities.  Currently, UYVY, NV12 and Y8C8 
 * supported.
 *
 * UYVY - YUV 422 interleaved corresponding to V4L2_PIX_FMT_UYVY in v4l2
 * Y8C8 - YUV 422 semi planar. The dm6467 VDCE outputs this format after a
 *        color conversion.The format consists of two planes: one with the
 *        Y component and one with the CbCr components interleaved (hence semi)  *
 *        See the LSP VDCE documentation for a thorough description of this
 *        format.
 * NV12 - 8-bit Y plane followed by an interleaved U/V plane with 2x2 
 *        subsampling
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS
    ("video/x-raw-yuv, "                        /* UYVY - YUV422 interleaved */
         "format=(fourcc)UYVY, "                
         "framerate=(fraction)[ 0, MAX ], "
         "width=(int)[ 1, MAX ], "
         "height=(int)[ 1, MAX ];"
     "video/x-raw-yuv, "                        /* Y8C8 - YUV422 semi planar */
         "format=(fourcc)Y8C8, "               
         "framerate=(fraction)[ 0, MAX ], "
         "width=(int)[ 1, MAX ], "
         "height=(int)[ 1, MAX ];"
    "video/x-raw-yuv, "                         /* NV16 - YUV422 semi planar */
         "format=(fourcc)NV16, "
         "framerate=(fraction)[ 0, MAX ], "
         "width=(int)[ 1, MAX ], "
         "height=(int)[ 1, MAX ];"
     "video/x-raw-yuv, "                        /* NV12 - YUV420 semi planar */
         "format=(fourcc)NV12, "               
         "framerate=(fraction)[ 0, MAX ], "
         "width=(int)[ 1, MAX ], "
         "height=(int)[ 1, MAX ]"
    )
);

/* Declare a global pointer to our element base class */
static GstElementClass *parent_class = NULL;

/* Static Function Declarations */
static void
 gst_tividenc1_base_init(gpointer g_class);
static void
 gst_tividenc1_class_init(GstTIVidenc1Class *g_class);
static void
 gst_tividenc1_init(GstTIVidenc1 *object, GstTIVidenc1Class *g_class);
static void
 gst_tividenc1_set_property (GObject *object, guint prop_id,
     const GValue *value, GParamSpec *pspec);
static void
 gst_tividenc1_get_property (GObject *object, guint prop_id, GValue *value,
     GParamSpec *pspec);
static gboolean
 gst_tividenc1_set_sink_caps(GstPad *pad, GstCaps *caps);
static gboolean
 gst_tividenc1_set_source_caps(GstTIVidenc1 *videnc1, Buffer_Handle hBuf);
static gboolean
 gst_tividenc1_sink_event(GstPad *pad, GstEvent *event);
static GstFlowReturn
 gst_tividenc1_chain(GstPad *pad, GstBuffer *buf);
static gboolean
 gst_tividenc1_init_video(GstTIVidenc1 *videnc1);
static gboolean
 gst_tividenc1_exit_video(GstTIVidenc1 *videnc1);
static GstStateChangeReturn
 gst_tividenc1_change_state(GstElement *element, GstStateChange transition);
static void*
 gst_tividenc1_encode_thread(void *arg);
static void
 gst_tividenc1_drain_pipeline(GstTIVidenc1 *videnc1);
static GstClockTime
 gst_tividenc1_frame_duration(GstTIVidenc1 *videnc1);
static ColorSpace_Type 
 gst_tividenc1_find_colorSpace (const gchar *colorSpace);
static gboolean
 gst_tividenc1_codec_start (GstTIVidenc1 *videnc1);
static gboolean
 gst_tividenc1_codec_stop (GstTIVidenc1 *videnc1);

/******************************************************************************
 * gst_tividenc1_class_init_trampoline
 *    Boiler-plate function auto-generated by "make_element" script.
 ******************************************************************************/
static void gst_tividenc1_class_init_trampoline(gpointer g_class, gpointer data)
{
    parent_class = (GstElementClass*) g_type_class_peek_parent(g_class);
    gst_tividenc1_class_init((GstTIVidenc1Class*)g_class);
}

/******************************************************************************
 * gst_tividenc1_get_type
 *    Boiler-plate function auto-generated by "make_element" script.
 *    Defines function pointers for initialization routines for this element.
 ******************************************************************************/
GType gst_tividenc1_get_type(void)
{
    static GType object_type = 0;

    if (G_UNLIKELY(object_type == 0)) {
        static const GTypeInfo object_info = {
            sizeof(GstTIVidenc1Class),
            gst_tividenc1_base_init,
            NULL,
            gst_tividenc1_class_init_trampoline,
            NULL,
            NULL,
            sizeof(GstTIVidenc1),
            0,
            (GInstanceInitFunc) gst_tividenc1_init
        };

        object_type = g_type_register_static((gst_element_get_type()),
                          "GstTIVidenc1", &object_info, (GTypeFlags)0);

        /* Initialize GST_LOG for this object */
        GST_DEBUG_CATEGORY_INIT(gst_tividenc1_debug, "TIVidenc1", 0,
            "TI xDM 1.x Video Encoder");

        GST_LOG("initialized get_type\n");
    }

    return object_type;
};

/******************************************************************************
 * gst_tividenc1_base_init
 *    Boiler-plate function auto-generated by "make_element" script.
 *    Initializes element base class.
 ******************************************************************************/
static void gst_tividenc1_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "TI xDM 1.x Video Encoder",
        "Codec/Encoder/Video",
        "Encodes video using an xDM 1.x-based codec",
        "Brijesh Singh; Texas Instruments, Inc."
    };

    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(element_class,
        gst_static_pad_template_get (&src_factory));
    gst_element_class_add_pad_template(element_class,
        gst_static_pad_template_get (&sink_factory));
    gst_element_class_set_details(element_class, &element_details);

}

/******************************************************************************
 * gst_tividenc1_class_init
 *    Boiler-plate function auto-generated by "make_element" script.
 *    Initializes the TIVidenc1 class.
 ******************************************************************************/
static void gst_tividenc1_class_init(GstTIVidenc1Class *klass)
{
    GObjectClass    *gobject_class;
    GstElementClass *gstelement_class;

    gobject_class    = (GObjectClass*)    klass;
    gstelement_class = (GstElementClass*) klass;

    gobject_class->set_property = gst_tividenc1_set_property;
    gobject_class->get_property = gst_tividenc1_get_property;

    gstelement_class->change_state = gst_tividenc1_change_state;

    g_object_class_install_property(gobject_class, PROP_ENGINE_NAME,
        g_param_spec_string("engineName", "Engine Name",
            "Engine name used by Codec Engine", "unspecified",
            G_PARAM_READWRITE));

    g_object_class_install_property(gobject_class, PROP_CODEC_NAME,
        g_param_spec_string("codecName", "Codec Name", "Name of video codec",
            "unspecified", G_PARAM_READWRITE));

    g_object_class_install_property(gobject_class, PROP_RESOLUTION,
        g_param_spec_string("resolution", "Input resolution", 
            "Input resolution for the input video ('width''x''height')",
            "720x480", G_PARAM_READWRITE));

    g_object_class_install_property(gobject_class, PROP_IN_COLORSPACE,
        g_param_spec_string("iColorSpace", "Input colorspace",
            "Input color space (UYVY, Y8C8, NV16 or NV12)",
            "unspecified", G_PARAM_READWRITE));

    g_object_class_install_property(gobject_class, PROP_BITRATE,
        g_param_spec_int("bitRate",
            "encoder bitrate",
            "Set the video encoder bit rate",
            1, G_MAXINT32, DEFAULT_BIT_RATE, G_PARAM_WRITABLE));

    /* Number of input buffers in the circular queue.
     * This is the buffer that is recieved from the upstream 
     */
    g_object_class_install_property(gobject_class, PROP_NUM_INPUT_BUFS,
        g_param_spec_int("numInputBufs",
            "Number of Input buffers",
            "Number of input buffers in circular queue",
            1, G_MAXINT32, 1, G_PARAM_WRITABLE));

    /* We allow more than three output buffer because this is the buffer that
     * is sent to the downstream element.  It may be that we need to have
     * more than 3 buffer if the downstream element doesn't give the buffer
     * back in time for the codec to use.
     */
    g_object_class_install_property(gobject_class, PROP_NUM_OUTPUT_BUFS,
        g_param_spec_int("numOutputBufs",
            "Number of Ouput Buffers",
            "Number of output buffers to allocate for codec",
            3, G_MAXINT32, 3, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_RATE_CTRL_PRESET,
        g_param_spec_int("rateControlPreset",
            "Rate control",
            "Bit rate control"
            "\n\t\t\t1 - No rate control"
            "\n\t\t\t2 - Constant bit rate (CBR)"
            "\n\t\t\t3 - Variable bit rate (VBR)",
            1, G_MAXINT32, 1, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_ENCODING_PRESET,
        g_param_spec_int("encodingPreset",
            "Encoding preset",
            "Algorithm specific creation time parameters"
            "\n\t\t\t1 - Use codec default"
            "\n\t\t\t2 - High quality"
            "\n\t\t\t3 - High speed",
            1, G_MAXINT32, 1, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_FRAMERATE,
        g_param_spec_int("frameRate",
            "Frame rate to play output",
            "Communicate this framerate to downstream elements.  The frame "
            "rate specified should be an integer.  If 29.97fps is desired, "
            "specify 30 for this setting",
            1, G_MAXINT32, 30, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_DISPLAY_BUFFER,
        g_param_spec_boolean("displayBuffer", "Display Buffer",
            "Display circular buffer status while processing",
            FALSE, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_BYTE_STREAM,
        g_param_spec_boolean("byteStream", "byte stream",
            "Generate byte stream format of NALU",
            TRUE, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_CONTIG_INPUT_BUF,
        g_param_spec_boolean("contiguousInputFrame", "Contiguous Input frame",
            "Set this if elemenet recieved contiguous input frame",
            FALSE, G_PARAM_WRITABLE));

    g_object_class_install_property(gobject_class, PROP_GEN_TIMESTAMPS,
        g_param_spec_boolean("genTimeStamps", "Generate Time Stamps",
            "Set timestamps on output buffers",
            TRUE, G_PARAM_WRITABLE));
}

/******************************************************************************
 * gst_tividenc1_init
 *    Initializes a new element instance, instantiates pads and sets the pad
 *    callback functions.
 ******************************************************************************/
static void gst_tividenc1_init(GstTIVidenc1 *videnc1, GstTIVidenc1Class *gclass)
{
    /* Instantiate encoded video sink pad.
     *
     * Fixate on our static template caps instead of writing a getcaps
     * function, which is overkill for this element.
     */
    videnc1->sinkpad =
        gst_pad_new_from_static_template(&sink_factory, "sink");
    gst_pad_set_setcaps_function(
        videnc1->sinkpad, GST_DEBUG_FUNCPTR(gst_tividenc1_set_sink_caps));
    gst_pad_set_event_function(
        videnc1->sinkpad, GST_DEBUG_FUNCPTR(gst_tividenc1_sink_event));
    gst_pad_set_chain_function(
        videnc1->sinkpad, GST_DEBUG_FUNCPTR(gst_tividenc1_chain));
    gst_pad_fixate_caps(videnc1->sinkpad,
        gst_caps_make_writable(
            gst_caps_copy(gst_pad_get_pad_template_caps(videnc1->sinkpad))));

    /* Instantiate deceoded video source pad.
     *
     * Fixate on our static template caps instead of writing a getcaps
     * function, which is overkill for this element.
     */
    videnc1->srcpad =
        gst_pad_new_from_static_template(&src_factory, "src");
    gst_pad_fixate_caps(videnc1->srcpad,
        gst_caps_make_writable(
            gst_caps_copy(gst_pad_get_pad_template_caps(videnc1->srcpad))));

    /* Add pads to TIVidenc1 element */
    gst_element_add_pad(GST_ELEMENT(videnc1), videnc1->sinkpad);
    gst_element_add_pad(GST_ELEMENT(videnc1), videnc1->srcpad);

    /* Determine target board type */
    if (Cpu_getDevice(NULL, &videnc1->device) < 0) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
        ("Failed to determine target board\n"), (NULL));
        videnc1->device = INVALID_DEVICE;
    }

    /* Initialize TIVidenc1 state */
    videnc1->engineName             = NULL;
    videnc1->codecName              = NULL;
    videnc1->displayBuffer          = FALSE;
    videnc1->genTimeStamps          = TRUE;

    videnc1->hEngine                = NULL;
    videnc1->hVe1                   = NULL;
    videnc1->drainingEOS            = FALSE;
    videnc1->threadStatus           = 0UL;

    videnc1->waitOnEncodeThread     = NULL;
    videnc1->waitOnEncodeDrain      = NULL;

    videnc1->framerateNum           = 0;
    videnc1->framerateDen           = 0;
    videnc1->hOutBufTab             = NULL;
    videnc1->circBuf                = NULL;

    videnc1->width                  = 0;
    videnc1->height                 = 0;
    videnc1->bitRate                = -1;
    videnc1->colorSpace             = ColorSpace_NOTSET;

    videnc1->numOutputBufs          = 0;
    videnc1->numInputBufs           = 0;
    videnc1->upstreamBufSize        = -1;
    videnc1->hCcv                   = NULL;
    videnc1->hFc                    = NULL;
    videnc1->rateControlPreset      = 1;
    videnc1->contiguousInputFrame   = FALSE;
    videnc1->encodingPreset         = 1;
    videnc1->byteStream             = TRUE;
    videnc1->codec_data             = NULL;
}

/******************************************************************************
 * gst_tividenc1_find_colorSpace
 *****************************************************************************/
static ColorSpace_Type gst_tividenc1_find_colorSpace (const gchar *colorSpace)
{
    if (!strcmp(colorSpace, "UYVY"))
        return ColorSpace_UYVY;
    else if (!strcmp(colorSpace, "NV16")) 
        return ColorSpace_YUV422PSEMI;
    else if (!strcmp(colorSpace, "Y8C8")) 
        return ColorSpace_YUV422PSEMI;
    else if (!strcmp(colorSpace, "NV12")) 
        return ColorSpace_YUV420PSEMI;
    else 
        return ColorSpace_NOTSET; 
}

/******************************************************************************
 * gst_tividenc1_set_property
 *     Set element properties when requested.
 ******************************************************************************/
static void gst_tividenc1_set_property(GObject *object, guint prop_id,
                const GValue *value, GParamSpec *pspec)
{
    GstTIVidenc1 *videnc1 = GST_TIVIDENC1(object);

    GST_LOG("begin set_property\n");

    switch (prop_id) {
        case PROP_IN_COLORSPACE:
            if (videnc1->iColor) {
                g_free((gpointer)videnc1->iColor);
            }
            videnc1->iColor =
                (gchar*)g_malloc(strlen(g_value_get_string(value)) + 1);
            strcpy((gchar *)videnc1->iColor, g_value_get_string(value));
            videnc1->colorSpace =
                 gst_tividenc1_find_colorSpace(g_value_get_string(value));
            GST_LOG("setting \"iColorSpace\" to \"%d\"\n",videnc1->colorSpace);
            break; 
        case PROP_RESOLUTION:
            if (videnc1->resolution) {
                g_free((gpointer)videnc1->resolution);
            }
            videnc1->resolution =
                (gchar*)g_malloc(strlen(g_value_get_string(value)) + 1);
            strcpy((gchar *)videnc1->resolution, g_value_get_string(value));
            sscanf(g_value_get_string(value), "%dx%d", &videnc1->width,
                        &videnc1->height);
            GST_LOG("setting \"resolution\" to \"%dx%d\"\n", videnc1->width,
                        videnc1->height);
            break;
        case PROP_BITRATE:
            videnc1->bitRate =  g_value_get_int(value);
            GST_LOG("setting \"bitRate\" to \"%d\" \n", videnc1->bitRate);
            break; 
        case PROP_ENGINE_NAME:
            if (videnc1->engineName) {
                g_free((gpointer)videnc1->engineName);
            }
            videnc1->engineName =
                (gchar*)g_malloc(strlen(g_value_get_string(value)) + 1);
            strcpy((gchar *)videnc1->engineName, g_value_get_string(value));
            GST_LOG("setting \"engineName\" to \"%s\"\n", videnc1->engineName);
            break;
        case PROP_CODEC_NAME:
            if (videnc1->codecName) {
                g_free((gpointer)videnc1->codecName);
            }
            videnc1->codecName =
                (gchar*)g_malloc(strlen(g_value_get_string(value)) + 1);
            strcpy((gchar*)videnc1->codecName, g_value_get_string(value));
            GST_LOG("setting \"codecName\" to \"%s\"\n", videnc1->codecName);
            break;
        case PROP_NUM_OUTPUT_BUFS:
            videnc1->numOutputBufs = g_value_get_int(value);
            GST_LOG("setting \"numOutputBufs\" to \"%d\" \n",
                     videnc1->numOutputBufs);
            break;
        case PROP_RATE_CTRL_PRESET:
            videnc1->rateControlPreset = g_value_get_int(value);
            GST_LOG("setting \"rateControlPreset\" to \"%d\" \n",
                     videnc1->rateControlPreset);
            break;
        case PROP_ENCODING_PRESET:
            videnc1->encodingPreset = g_value_get_int(value);
            GST_LOG("setting \"encodingPreset\" to \"%d\" \n",
                     videnc1->encodingPreset);
            break;
        case PROP_NUM_INPUT_BUFS:
            videnc1->numInputBufs = g_value_get_int(value);
            GST_LOG("setting \"numInputBufs\" to \"%d\" \n",
                     videnc1->numInputBufs);
            break;
        case PROP_FRAMERATE:
        {
            videnc1->framerateNum = g_value_get_int(value);
            videnc1->framerateDen = 1;

            /* If 30fps was specified, use 29.97 */
            if (videnc1->framerateNum == 30) {
                videnc1->framerateNum = 30000;
                videnc1->framerateDen = 1001;
            }

            GST_LOG("setting \"frameRate\" to \"%2.2lf\"\n",
               (gdouble)videnc1->framerateNum / (gdouble)videnc1->framerateDen);
            break;
        }
        case PROP_CONTIG_INPUT_BUF:
            videnc1->contiguousInputFrame = g_value_get_boolean(value);
            GST_LOG("setting \"contiguousInputFrame\" to \"%s\"\n",
                videnc1->contiguousInputFrame ? "TRUE" : "FALSE");
            break;
        case PROP_DISPLAY_BUFFER:
            videnc1->displayBuffer = g_value_get_boolean(value);
            GST_LOG("setting \"displayBuffer\" to \"%s\"\n",
                videnc1->displayBuffer ? "TRUE" : "FALSE");
            break;
        case PROP_BYTE_STREAM:
            videnc1->byteStream = g_value_get_boolean(value);
            GST_LOG("setting \"byteStream\" to \"%s\"\n",
                videnc1->byteStream ? "TRUE" : "FALSE");
            break;
        case PROP_GEN_TIMESTAMPS:
            videnc1->genTimeStamps = g_value_get_boolean(value);
            GST_LOG("setting \"genTimeStamps\" to \"%s\"\n",
                videnc1->genTimeStamps ? "TRUE" : "FALSE");
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
            break;
    }

    GST_LOG("end set_property\n");
}

/******************************************************************************
 * gst_tividenc1_get_property
 *     Return values for requested element property.
 ******************************************************************************/
static void gst_tividenc1_get_property(GObject *object, guint prop_id,
                GValue *value, GParamSpec *pspec)
{
    GstTIVidenc1 *videnc1 = GST_TIVIDENC1(object);

    GST_LOG("begin get_property\n");

    switch (prop_id) {
        case PROP_ENGINE_NAME:
            g_value_set_string(value, videnc1->engineName);
            break;
        case PROP_CODEC_NAME:
            g_value_set_string(value, videnc1->codecName);
            break;
        case PROP_RESOLUTION:
            g_value_set_string(value, videnc1->resolution);
            break;
        case PROP_IN_COLORSPACE:
            g_value_set_string(value, videnc1->iColor);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
            break;
    }

    GST_LOG("end get_property\n");
}


/******************************************************************************
 * gst_tividenc1_set_sink_caps
 *     Negotiate our sink pad capabilities.
 ******************************************************************************/
static gboolean gst_tividenc1_set_sink_caps(GstPad *pad, GstCaps *caps)
{
    GstTIVidenc1  *videnc1;
    GstStructure *capStruct;
    const gchar  *mime;
    char         *string;

    videnc1    = GST_TIVIDENC1(gst_pad_get_parent(pad));
    capStruct = gst_caps_get_structure(caps, 0);
    mime      = gst_structure_get_name(capStruct);

    string = gst_caps_to_string(caps);
    GST_INFO("requested sink caps:  %s", string);
    g_free(string);

    /* Shut-down any running video encoder */
    if (!gst_tividenc1_exit_video(videnc1)) {
        gst_object_unref(videnc1);
        return FALSE;
    }

    /* Generic Video Properties */
    if (!strncmp(mime, "video/", 6)) {
        gint  framerateNum;
        gint  framerateDen;

        if (gst_structure_get_fraction(capStruct, "framerate", &framerateNum,
                &framerateDen)) {
            videnc1->framerateNum = framerateNum;
            videnc1->framerateDen = framerateDen;
        }
    }

    /* RAW YUV input */
    if (!strcmp(mime,"video/x-raw-yuv")) {
        guint32 fourcc;

        if (gst_structure_get_fourcc(capStruct, "format", &fourcc)) {

            switch (fourcc) {
                case GST_MAKE_FOURCC('U', 'Y', 'V', 'Y'):
                    videnc1->colorSpace = ColorSpace_UYVY;
                    break;

                case GST_MAKE_FOURCC('N', 'V', '1', '6'):
                case GST_MAKE_FOURCC('Y', '8', 'C', '8'):
                    videnc1->colorSpace = ColorSpace_YUV422PSEMI;
                    break;

                case GST_MAKE_FOURCC('N', 'V', '1', '2'):
                    videnc1->colorSpace = ColorSpace_YUV420PSEMI;
                    break;

                default:
                    GST_ELEMENT_ERROR(videnc1, STREAM, NOT_IMPLEMENTED,
                    ("unsupported fourcc in video stream\n"), (NULL));
                    gst_object_unref(videnc1);
                    return FALSE;
            }

            if (videnc1->width == 0 && 
               !gst_structure_get_int(capStruct, "width", &videnc1->width)) {
                videnc1->width = 0;
            }

            if (videnc1->height == 0 &&
               !gst_structure_get_int(capStruct, "height", &videnc1->height)) {
                videnc1->height = 0;
            }
        }
    }

    /* Mime type not supported */
    else {
        GST_ELEMENT_ERROR(videnc1, STREAM, NOT_IMPLEMENTED,
        ("stream type not supported"), (NULL));
        gst_object_unref(videnc1);
        return FALSE;
    }

    gst_object_unref(videnc1);

    GST_LOG("sink caps negotiation successful\n");
    return TRUE;
}

/******************************************************************************
 * gst_tividenc1_set_source_caps
 *     Negotiate our source pad capabilities.
 ******************************************************************************/
static gboolean gst_tividenc1_set_source_caps(
                    GstTIVidenc1 *videnc1, Buffer_Handle hBuf)
{
    BufferGfx_Dimensions  dim;
    GstCaps              *caps = NULL;
    gboolean              ret;
    GstPad               *pad;
    char                 *string;
    GstTICodec           *h264Codec  = NULL;
    GstTICodec           *mpeg4Codec = NULL;
    GstTICodec           *mpeg2Codec = NULL;

    h264Codec  =  gst_ticodec_get_codec("H.264 Video Encoder");
    mpeg4Codec  =  gst_ticodec_get_codec("MPEG4 Video Encoder");
    mpeg2Codec  =  gst_ticodec_get_codec("MPEG2 Video Encoder");

    pad = videnc1->srcpad;

    /* Create a video caps object using the dimensions from the given buffer */
    BufferGfx_getDimensions(hBuf, &dim);

    /* Create H.264 source cap */
    if (h264Codec && (!strcmp(h264Codec->CE_CodecName, videnc1->codecName))) {
        caps =
            gst_caps_new_simple("video/x-h264",
                "framerate",    GST_TYPE_FRACTION,  videnc1->framerateNum,
                                                    videnc1->framerateDen,
                "width",        G_TYPE_INT,         dim.width,
                "height",       G_TYPE_INT,         dim.height,
                NULL);
   
        if (videnc1->byteStream == FALSE) {
            gst_caps_set_simple(caps, "codec_data", GST_TYPE_BUFFER, 
                videnc1->codec_data, (char*)NULL);
        }
     
        string =  gst_caps_to_string(caps);
        GST_LOG("setting source caps to x-h264: %s", string);
        g_free(string);
    }
 
    /* Create MPEG-2 source cap */
    if (mpeg2Codec && (!strcmp(mpeg2Codec->CE_CodecName, videnc1->codecName))) {
        gint mpegversion = 2;

        caps =
            gst_caps_new_simple("video/mpeg",
                "mpegversion",  G_TYPE_INT,         mpegversion,
                "framerate",    GST_TYPE_FRACTION,  videnc1->framerateNum,
                                                    videnc1->framerateDen,
                "width",        G_TYPE_INT,         dim.width,
                "height",       G_TYPE_INT,         dim.height,
                NULL);

        string =  gst_caps_to_string(caps); 
        GST_LOG("setting source caps to mpeg2: %s", string);
        g_free(string);
    }

    /* Create MPEG4 source cap */
    if (mpeg4Codec && (!strcmp(mpeg4Codec->CE_CodecName, videnc1->codecName))) {
        gint mpegversion = 4;

        caps =
            gst_caps_new_simple("video/mpeg",
                "mpegversion",  G_TYPE_INT,         mpegversion,
                "framerate",    GST_TYPE_FRACTION,  videnc1->framerateNum,
                                                    videnc1->framerateDen,
                "width",        G_TYPE_INT,         dim.width,
                "height",       G_TYPE_INT,         dim.height,
                NULL);

        string =  gst_caps_to_string(caps); 
        GST_LOG("setting source caps to mpeg4: %s", string);
        g_free(string);
    }

    /* Set the source pad caps */
    ret = gst_pad_set_caps(pad, caps);
    gst_caps_unref(caps);

    return ret;
}

/******************************************************************************
 * gst_tividenc1_sink_event
 *     Perform event processing on the input stream.  At the moment, this
 *     function isn't needed as this element doesn't currently perform any
 *     specialized event processing.  We'll leave it in for now in case we need
 *     it later on.
 ******************************************************************************/
static gboolean gst_tividenc1_sink_event(GstPad *pad, GstEvent *event)
{
    GstTIVidenc1 *videnc1;
    gboolean     ret;

    videnc1 = GST_TIVIDENC1(GST_OBJECT_PARENT(pad));

    GST_DEBUG("pad \"%s\" received:  %s\n", GST_PAD_NAME(pad),
        GST_EVENT_TYPE_NAME(event));

    switch (GST_EVENT_TYPE(event)) {

        case GST_EVENT_NEWSEGMENT:
            /* maybe save and/or update the current segment (e.g. for output
             * clipping) or convert the event into one in a different format
             * (e.g. BYTES to TIME) or drop it and set a flag to send a
             * newsegment event in a different format later
             */
            ret = gst_pad_push_event(videnc1->srcpad, event);
            break;

        case GST_EVENT_EOS:
            /* end-of-stream: process any remaining encoded frame data */
            GST_LOG("no more input; draining remaining encoded video data\n");

            if (!videnc1->drainingEOS) {
               gst_tividenc1_drain_pipeline(videnc1);
             }

            /* Propagate EOS to downstream elements */
            ret = gst_pad_push_event(videnc1->srcpad, event);
            break;

        case GST_EVENT_FLUSH_STOP:
            ret = gst_pad_push_event(videnc1->srcpad, event);
            break;

        /* Unhandled events */
        case GST_EVENT_BUFFERSIZE:
        case GST_EVENT_CUSTOM_BOTH:
        case GST_EVENT_CUSTOM_BOTH_OOB:
        case GST_EVENT_CUSTOM_DOWNSTREAM:
        case GST_EVENT_CUSTOM_DOWNSTREAM_OOB:
        case GST_EVENT_CUSTOM_UPSTREAM:
        case GST_EVENT_FLUSH_START:
        case GST_EVENT_NAVIGATION:
        case GST_EVENT_QOS:
        case GST_EVENT_SEEK:
        case GST_EVENT_TAG:
        default:
            ret = gst_pad_event_default(pad, event);
            break;

    }

    return ret;
}

/******************************************************************************
 * gst_tividenc1_convert_gst_to_dmai
 *  This function convert gstreamer buffer into DMAI graphics buffer.
 *****************************************************************************/
static Buffer_Handle gst_tividenc1_convert_gst_to_dmai(GstTIVidenc1 *videnc1, 
    GstBuffer *buf, gboolean reference)
{
    BufferGfx_Attrs gfxAttrs   = BufferGfx_Attrs_DEFAULT;
    Buffer_Handle   hBuf       = NULL;
 
    gfxAttrs.bAttrs.reference  = reference;
    gfxAttrs.dim.width         = videnc1->width;
    gfxAttrs.dim.height        = videnc1->height;
    gfxAttrs.colorSpace        = videnc1->colorSpace;
    gfxAttrs.dim.lineLength    = BufferGfx_calcLineLength(
                                    gfxAttrs.dim.width, 
                                     videnc1->colorSpace);
    hBuf = Buffer_create(GST_BUFFER_SIZE(buf),
                BufferGfx_getBufferAttrs(&gfxAttrs));
    if (hBuf == NULL) {
        GST_ERROR("failed to create  buffer\n");
        return NULL;
    }
    Buffer_setUserPtr(hBuf, (Int8*)GST_BUFFER_DATA(buf));
    Buffer_setNumBytesUsed(hBuf, GST_BUFFER_SIZE(buf));

    return hBuf;
}

/******************************************************************************
 * gst_tividenc1_422psemi_420psemi
 *  this function color convert YUV422PSEMI to YUV420PSEMI.
 *****************************************************************************/
static Int gst_tividenc1_422psemi_420psemi(Int8* dst, GstBuffer *src, 
    GstTIVidenc1   *videnc1)
{
    BufferGfx_Attrs gfxAttrs    = BufferGfx_Attrs_DEFAULT;
    Buffer_Handle  hInBuf       = NULL;
    Buffer_Handle  hOutBuf      = NULL;
    Int ret                     = -1;
    Ccv_Attrs       ccvAttrs    = Ccv_Attrs_DEFAULT;
    gboolean        accel       = FALSE;

    GST_LOG("gst_tividenc1_422psemi_420psemi - begin\n");

    /* create ccv handle */
    if (videnc1->hCcv == NULL) {
        /* Enable the accel ccv based on contiguousInputFrame.
         * If accel is set to FALSE then DMAI will use software ccv function
         * else will use HW accelerated ccv engine.
         */

        /* If we are getting dmai transport buffer then enable HW 
         * acceleration */
        if (GST_IS_TIDMAIBUFFERTRANSPORT(src)) {
            accel = TRUE;
        }
        else {
            accel = videnc1->contiguousInputFrame;
        }

        ccvAttrs.accel = videnc1->contiguousInputFrame;
        videnc1->hCcv = Ccv_create(&ccvAttrs);
        if (videnc1->hCcv == NULL) {
            GST_ERROR("failed to create CCV handle\n");
            goto exit;
        }
        
        GST_INFO("HW accel CCV: %s\n", accel ? "enabled":"disabled");
    }

    /* Prepare input buffer */
    hInBuf = gst_tividenc1_convert_gst_to_dmai(videnc1, src, TRUE);
    if (hInBuf == NULL) {
        GST_ERROR("failed to get dmai buffer\n");
        goto exit;
    }

    /* Prepare output buffer */
    gfxAttrs.bAttrs.reference = TRUE;
    gfxAttrs.colorSpace = ColorSpace_YUV420PSEMI;
    gfxAttrs.dim.width  = videnc1->width;
    gfxAttrs.dim.height = videnc1->height;
    gfxAttrs.dim.lineLength = BufferGfx_calcLineLength(videnc1->width, 
                               videnc1->colorSpace);

    hOutBuf = Buffer_create(gfxAttrs.dim.lineLength * videnc1->height * 3 / 2,
                BufferGfx_getBufferAttrs(&gfxAttrs));
    if (hOutBuf == NULL) {
        GST_ERROR("failed to create dest buffer\n");
        goto exit;
    }
    Buffer_setUserPtr(hOutBuf, dst);

    if (Ccv_config(videnc1->hCcv, hInBuf, hOutBuf) < 0) {
        GST_ERROR("failed to config CCV handle\n");
        goto exit;
    }

    if (Ccv_execute(videnc1->hCcv, hInBuf, hOutBuf) < 0) {
        GST_ERROR("failed to execute Ccv handle\n");
        goto exit;
    }

    ret = GST_BUFFER_SIZE(src);

    GST_LOG("gst_tividenc1_422psemi_420psemi - end\n");
exit:
    if (hInBuf) {
        Buffer_delete(hInBuf);
    }

    if (hOutBuf) {
        Buffer_delete(hOutBuf);
    }

    return ret;
}

/*****************************************************************************
 * gst_tividenc1_circbuf_framecopy
 *  This function will be invoked by circular buffer during copy method.
 *  Function performs copy from source to destination buffer using
 *  DMAI framecopy module.
 ****************************************************************************/
static Int gst_tividenc1_circbuf_copy (Int8 *dst, GstBuffer *src, void *data)
{
    GstTIVidenc1   *videnc1     = (GstTIVidenc1*) data;
    BufferGfx_Attrs gfxAttrs    = BufferGfx_Attrs_DEFAULT;
    Buffer_Handle  hInBuf       = NULL;
    Buffer_Handle  hOutBuf      = NULL;
    Int ret                     = -1;
    Framecopy_Attrs fcAttrs     = Framecopy_Attrs_DEFAULT;
    gboolean        accel       = FALSE;

    #if defined(Platform_dm365)
    BufferGfx_Dimensions  dim;
    #endif

    /* If we are not recieving full frame then use memcpy() and let circular
     * manage everything else.
     */
    if (GST_BUFFER_SIZE(src) < gst_ti_calc_buffer_size(videnc1->width,
         videnc1->height, 0, videnc1->colorSpace)) {
        memcpy(dst, GST_BUFFER_DATA(src), GST_BUFFER_SIZE(src));
        return GST_BUFFER_SIZE(src);
    }
 
    /* Check to see if we need to execute ccv on dm6467 */
    if (videnc1->device == Cpu_Device_DM6467 &&
         videnc1->colorSpace == ColorSpace_YUV422PSEMI) {
        return gst_tividenc1_422psemi_420psemi(dst, src, videnc1);
    }

    /* TODO:
     * DMAI 1.20.00.06 has BUG in non-accel framecopy for YUV420PSEMI. 
     * Use memcpy() to copy upstream buffer in circular buffer. And this 
     * workaround  *must* be removed once move to later version of DMAI for 
     * DM6467.
     */
    if (!(videnc1->contiguousInputFrame) &&
            (videnc1->device == Cpu_Device_DM6467) &&
             (videnc1->colorSpace == ColorSpace_YUV420PSEMI)) {

        memcpy(dst, GST_BUFFER_DATA(src), GST_BUFFER_SIZE(src));
        return GST_BUFFER_SIZE(src);
    }

    GST_LOG("gst_tividenc1_circbuf_framecopy - begin\n");

    if (videnc1->hFc == NULL) {
        /* Enable the accel framecopy based on contiguousInputFrame.
         * If accel is set to FALSE then DMAI will use regular memcpy function
         * else will use HW accelerated framecopy.
         */

        /* If we are getting dmai transport buffer then enable HW 
         * acceleration */
        if (GST_IS_TIDMAIBUFFERTRANSPORT(src)) {
            accel = TRUE;
        }
        else {
            accel = videnc1->contiguousInputFrame;
        }

        fcAttrs.accel = videnc1->contiguousInputFrame;

        videnc1->hFc = Framecopy_create(&fcAttrs);
        if (videnc1->hFc == NULL) {
            GST_ERROR("failed to create framecopy handle\n");
            goto exit;
        }

        GST_INFO("HW accel framecopy: %s\n", accel ? "enabled":"disabled");
    }

    /* Prepare input buffer */
    hInBuf = gst_tividenc1_convert_gst_to_dmai(videnc1, src, TRUE);
    if (hInBuf == NULL) {
        GST_ERROR("failed to get dmai buffer\n");
        goto exit;
    }

    #if defined(Platform_dm365)
    /* Handle resizer 32-byte issue on DM365 platform */
    if (videnc1->device ==  Cpu_Device_DM365) {
        if ((videnc1->colorSpace ==  ColorSpace_YUV420PSEMI)) {
            BufferGfx_getDimensions(hInBuf, &dim);
            dim.lineLength = Dmai_roundUp(dim.lineLength,32);
            BufferGfx_setDimensions(hInBuf, &dim);
        }
    }
    #endif

    /* Prepare output buffer */
    gfxAttrs.bAttrs.reference = TRUE;
    gfxAttrs.colorSpace = videnc1->colorSpace;
    gfxAttrs.dim.width  = videnc1->width;
    gfxAttrs.dim.height = videnc1->height;
    gfxAttrs.dim.lineLength = BufferGfx_calcLineLength(videnc1->width, 
                               videnc1->colorSpace);

    hOutBuf = Buffer_create(GST_BUFFER_SIZE(src),
                BufferGfx_getBufferAttrs(&gfxAttrs));
    if (hOutBuf == NULL) {
        GST_ERROR("failed to create dest buffer\n");
        goto exit;
    }
    Buffer_setUserPtr(hOutBuf, dst);

    if (Framecopy_config(videnc1->hFc, hInBuf, hOutBuf) < 0) {
        GST_ERROR("failed to configure framecopy\n");
        goto exit;
    }

    if (Framecopy_execute(videnc1->hFc, hInBuf, hOutBuf) < 0) {
        GST_ERROR("failed to execute framecopy\n");
        goto exit;
    }

    ret = GST_BUFFER_SIZE(src);

exit:
    if (hInBuf) {
        Buffer_delete(hInBuf);
    }

    if (hOutBuf) {
        Buffer_delete(hOutBuf);
    }

    GST_LOG("gst_tividenc1_circbuf_framecopy - end\n");
    return ret;
}

/******************************************************************************
 * gst_tividenc1_chain
 *    This is the main processing routine.  This function receives a buffer
 *    from the sink pad, processes it, and pushes the result to the source
 *    pad.
 ******************************************************************************/
static GstFlowReturn gst_tividenc1_chain(GstPad * pad, GstBuffer * buf)
{
    GstTIVidenc1   *videnc1     = GST_TIVIDENC1(GST_OBJECT_PARENT(pad));
    GstFlowReturn   flow        = GST_FLOW_OK;
    gboolean        checkResult;

    /* If the encode thread aborted, signal it to let it know it's ok to
     * shut down, and communicate the failure to the pipeline.
     */
    if (gst_tithread_check_status(videnc1, TIThread_CODEC_ABORTED,
            checkResult)) {
        flow = GST_FLOW_UNEXPECTED;
        goto exit;
    }

    /* If our engine handle is currently NULL, then either this is our first
     * buffer or the upstream element has re-negotiated our capabilities which
     * resulted in our engine being closed.  In either case, we need to
     * initialize (or re-initialize) our video encoder to handle the new
     * stream.
     */
    if (videnc1->hEngine == NULL) {

        /* Calculate the input buffer size for circular buffer */
        /* DM6467: If we are recieving YUV422PSEMI from the upstream then create
         * big enough circular buffer to hold the YUV422PSEMI data. This is 
         * mainly because codec accepts YUV420PSEMI and we will execute
         * CCV in encode thread before calling video encoder.
         */
        if ((videnc1->device == Cpu_Device_DM6467) && 
              (videnc1->colorSpace == ColorSpace_YUV422PSEMI)) {
            videnc1->upstreamBufSize = GST_BUFFER_SIZE(buf);
        }
    
        /* DM365: If we are recieving YUV420PSEMI buffer from upstream
         * then we may need to componsate resizer 32-byte alignment issue. 
         * Hence create big enough circular buffer to hold the 32-byte aligned 
         * data from upstream.
         */
        #if defined(Platform_dm365)
        if ((videnc1->device == Cpu_Device_DM365) && 
             (videnc1->colorSpace == ColorSpace_YUV420PSEMI)) {
            videnc1->upstreamBufSize = GST_BUFFER_SIZE(buf);
        }
        #endif

        /* Initialize video encoder thread */
        if (!gst_tividenc1_init_video(videnc1)) {
            GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
            ("unable to initialize video\n"), (NULL));
            flow = GST_FLOW_UNEXPECTED;
            goto exit;
        }
            
        GST_TICIRCBUFFER_TIMESTAMP(videnc1->circBuf) =
            GST_CLOCK_TIME_IS_VALID(GST_BUFFER_TIMESTAMP(buf)) ?
            GST_BUFFER_TIMESTAMP(buf) : 0ULL;

        /* set circular buffer to use user defined copy function */
        if (gst_ticircbuffer_copy_config(videnc1->circBuf,  
            gst_tividenc1_circbuf_copy, (void*)videnc1) < 0) {
            GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
            ("failed to configure user defined copy\n"),(NULL));
            flow = GST_FLOW_UNEXPECTED;
            goto exit;
        }
    }

    /* Queue up the encoded data stream into a circular buffer */
    if (!gst_ticircbuffer_queue_data(videnc1->circBuf, buf)) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, WRITE,
        ("Failed to queue input buffer into circular buffer\n"), (NULL));
        flow = GST_FLOW_UNEXPECTED;
        goto exit;
    }

exit:
    gst_buffer_unref(buf);
    return flow;
}

/******************************************************************************
 * gst_tividenc1_init_video
 *     Initialize or re-initializes the video stream
 ******************************************************************************/
static gboolean gst_tividenc1_init_video(GstTIVidenc1 *videnc1)
{
    Rendezvous_Attrs    rzvAttrs = Rendezvous_Attrs_DEFAULT;
    struct sched_param  schedParam;
    pthread_attr_t      attr;

    GST_LOG("begin init_video\n");

    /* If video has already been initialized, shut down previous encoder */
    if (videnc1->hEngine) {
        if (!gst_tividenc1_exit_video(videnc1)) {
            GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
            ("failed to shut down existing video encoder\n"), (NULL));
            return FALSE;
        }
    }

    /* Make sure we know what codec we're using */
    if (!videnc1->engineName) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
        ("engine name not specified\n"), (NULL));
        return FALSE;
    }

    if (!videnc1->codecName) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
        ("codec name not specified\n"), (NULL));
        return FALSE;
    }

    /* Initialize thread status management */
    videnc1->threadStatus = 0UL;
    pthread_mutex_init(&videnc1->threadStatusMutex, NULL);

    /* Initialize rendezvous objects for making threads wait on conditions */
    videnc1->waitOnEncodeThread = Rendezvous_create(2, &rzvAttrs);
    videnc1->waitOnEncodeDrain  = Rendezvous_create(100, &rzvAttrs);
    videnc1->drainingEOS        = FALSE;

    /* Initialize custom thread attributes */
    if (pthread_attr_init(&attr)) {
        GST_WARNING("failed to initialize thread attrs\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Force the thread to use the system scope */
    if (pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM)) {
        GST_WARNING("failed to set scope attribute\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Force the thread to use custom scheduling attributes */
    if (pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED)) {
        GST_WARNING("failed to set schedule inheritance attribute\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Set the thread to be fifo real time scheduled */
    if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO)) {
        GST_WARNING("failed to set FIFO scheduling policy\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Set the display thread priority */
    schedParam.sched_priority = GstTIVideoThreadPriority;
    if (pthread_attr_setschedparam(&attr, &schedParam)) {
        GST_WARNING("failed to set scheduler parameters\n");
        return FALSE;
    }

    /* Create encoder thread */
    if (pthread_create(&videnc1->encodeThread, &attr,
            gst_tividenc1_encode_thread, (void*)videnc1)) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
        ("failed to create encode thread\n"), (NULL));
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }
    gst_tithread_set_status(videnc1, TIThread_CODEC_CREATED);

    /* Destroy the custom thread attributes */
    if (pthread_attr_destroy(&attr)) {
        GST_WARNING("failed to destroy thread attrs\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Make sure circular buffer and display buffer handles are created by
     * decoder thread.
     */
    Rendezvous_meet(videnc1->waitOnEncodeThread);

    if (videnc1->circBuf == NULL || videnc1->hOutBufTab == NULL) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
        ("encode thread failed to create circbuf or display buffer handles\n"),
        (NULL));
        return FALSE;
    }

    GST_LOG("end init_video\n");
    return TRUE;
}


/******************************************************************************
 * gst_tividenc1_exit_video
 *    Shut down any running video encoder, and reset the element state.
 ******************************************************************************/
static gboolean gst_tividenc1_exit_video(GstTIVidenc1 *videnc1)
{
    void*    thread_ret;
    gboolean checkResult;

    GST_LOG("begin exit_video\n");

    /* Drain the pipeline if it hasn't already been drained */
    if (!videnc1->drainingEOS) {
       gst_tividenc1_drain_pipeline(videnc1);
     }

    /* Shut down the encode thread */
    if (gst_tithread_check_status(
            videnc1, TIThread_CODEC_CREATED, checkResult)) {
        GST_LOG("shutting down encode thread\n");

        Rendezvous_force(videnc1->waitOnEncodeThread);
        if (pthread_join(videnc1->encodeThread, &thread_ret) == 0) {
            if (thread_ret == GstTIThreadFailure) {
                GST_DEBUG("encode thread exited with an error condition\n");
            }
        }
    }

    /* Shut down thread status management */
    videnc1->threadStatus = 0UL;
    pthread_mutex_destroy(&videnc1->threadStatusMutex);

    /* Shut down remaining items */
    if (videnc1->waitOnEncodeThread) {
        Rendezvous_delete(videnc1->waitOnEncodeThread);
        videnc1->waitOnEncodeThread = NULL;
    }

    if (videnc1->waitOnEncodeDrain) {
        Rendezvous_delete(videnc1->waitOnEncodeDrain);
        videnc1->waitOnEncodeDrain = NULL;
    }

    if (videnc1->hFc) {
        GST_LOG("freeing framecopy handle\n");
        Framecopy_delete(videnc1->hFc);
        videnc1->hFc = NULL;
    }

    if (videnc1->hCcv) {
        GST_LOG("freeing ccv handle\n");
        Ccv_delete(videnc1->hCcv);
        videnc1->hCcv = NULL;
    }

    if (videnc1->codec_data) {
        GST_LOG("freeing codec_data buffer\n");
        gst_buffer_unref(videnc1->codec_data);
        videnc1->codec_data = NULL;
    }

    GST_LOG("end exit_video\n");
    return TRUE;
}


/******************************************************************************
 * gst_tividenc1_change_state
 *     Manage state changes for the video stream.  The gStreamer documentation
 *     states that state changes must be handled in this manner:
 *        1) Handle ramp-up states
 *        2) Pass state change to base class
 *        3) Handle ramp-down states
 ******************************************************************************/
static GstStateChangeReturn gst_tividenc1_change_state(GstElement *element,
                                GstStateChange transition)
{
    GstStateChangeReturn  ret    = GST_STATE_CHANGE_SUCCESS;
    GstTIVidenc1          *videnc1 = GST_TIVIDENC1(element);

    GST_LOG("begin change_state (%d)\n", transition);

    /* Handle ramp-up state changes */
    switch (transition) {
        case GST_STATE_CHANGE_NULL_TO_READY:
            break;
        default:
            break;
    }

    /* Pass state changes to base class */
    ret = GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
    if (ret == GST_STATE_CHANGE_FAILURE)
        return ret;

    /* Handle ramp-down state changes */
    switch (transition) {
        case GST_STATE_CHANGE_READY_TO_NULL:
            /* Shut down any running video encoder */
            if (!gst_tividenc1_exit_video(videnc1)) {
                return GST_STATE_CHANGE_FAILURE;
            }
            break;

        default:
            break;
    }

    GST_LOG("end change_state\n");
    return ret;
}

/******************************************************************************
 * gst_tividenc1_codec_stop
 *   stop codec engine
 *****************************************************************************/
static gboolean gst_tividenc1_codec_stop (GstTIVidenc1 *videnc1)
{
    if (videnc1->circBuf) {
        GstTICircBuffer *circBuf;

        GST_LOG("freeing cicrular input buffer\n");

        circBuf               = videnc1->circBuf;
        videnc1->circBuf      = NULL;
        videnc1->framerateNum = 0;
        videnc1->framerateDen = 0;
        gst_ticircbuffer_unref(circBuf);
    }

    if (videnc1->hOutBufTab) {
        GST_LOG("freeing output buffers\n");
        gst_tidmaibuftab_unref(videnc1->hOutBufTab);
        videnc1->hOutBufTab = NULL;
    }

    if (videnc1->hVe1) {
        GST_LOG("closing video encoder\n");
        Venc1_delete(videnc1->hVe1);
        videnc1->hVe1 = NULL;
    }

    if (videnc1->hEngine) {
        GST_LOG("closing codec engine\n");
        Engine_close(videnc1->hEngine);
        videnc1->hEngine = NULL;
    }

    return TRUE;
}

/******************************************************************************
 * gst_tividenc1_codec_start
 *   start codec engine
 *****************************************************************************/
static gboolean gst_tividenc1_codec_start (GstTIVidenc1 *videnc1)
{
    VIDENC1_DynamicParams dynParams  = Venc1_DynamicParams_DEFAULT;
    BufferGfx_Attrs       gfxAttrs   = BufferGfx_Attrs_DEFAULT;
    Int                   inBufSize;
    VIDENC1_Params        params     = Venc1_Params_DEFAULT;

    /* Open the codec engine */
    GST_LOG("opening codec engine \"%s\"\n", videnc1->engineName);
    videnc1->hEngine = Engine_open((Char *) videnc1->engineName, NULL, NULL);

    if (videnc1->hEngine == NULL) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED,
        ("failed to open codec engine \"%s\"\n", videnc1->engineName), (NULL));
        return FALSE;
    }

    /* setup codec parameters depending on device */
    switch(videnc1->device) {
        case Cpu_Device_OMAP3530:
        #if defined(Platform_omapl138)
        case Cpu_Device_OMAPL138:
        #endif
        case Cpu_Device_DM6446:
            params.inputChromaFormat = XDM_YUV_422ILE;
            break;

        case Cpu_Device_DM355:
            params.inputChromaFormat = XDM_YUV_422ILE;
            params.reconChromaFormat = XDM_YUV_420P;
            break;

        case Cpu_Device_DM6467:
            #if defined(Platform_dm6467t)
            params.inputChromaFormat = XDM_YUV_420SP;
            #else
            params.inputChromaFormat = XDM_YUV_420P;
            #endif
            params.reconChromaFormat = XDM_CHROMA_NA;
            break;

        #if defined(Platform_dm365)
        case Cpu_Device_DM365:
            if (videnc1->colorSpace == ColorSpace_UYVY) {
                params.inputChromaFormat = XDM_YUV_422ILE;
            }
            else {
                params.inputChromaFormat = XDM_YUV_420SP;
                params.reconChromaFormat = XDM_YUV_420SP;
            }
            break;
        #endif

        default:
            GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED, 
             ("unsupport cpu type\n"), (NULL));
            return FALSE;
    }

    /* set the bit rate control preset */
    switch(videnc1->rateControlPreset) {
        case 1:
            params.rateControlPreset = IVIDEO_NONE;
            break;
        case 2:
            params.rateControlPreset = IVIDEO_LOW_DELAY;
            break;
        case 3:
            params.rateControlPreset = IVIDEO_STORAGE;
            break;
        default:
            params.rateControlPreset = IVIDEO_NONE;
            break;
    }
    
    /* Set the encoding preset */
    switch(videnc1->encodingPreset) {
        case 1:
            break;
        case 2:
            params.encodingPreset    = XDM_HIGH_QUALITY;
            break;
        case 3: 
            params.encodingPreset    = XDM_HIGH_SPEED;
            break;
        default:
            break;
    }

    /* set the encoding bitrate */
    params.maxBitRate = videnc1->bitRate < 0 ? 
        DEFAULT_BIT_RATE : videnc1->bitRate;

    params.maxWidth         = videnc1->width;
    params.maxHeight        = videnc1->height;
    dynParams.targetBitRate = params.maxBitRate;
    dynParams.inputWidth    = videnc1->width;
    dynParams.inputHeight   = videnc1->height;
    
    GST_LOG("configuring video encode width=%ld, height=%ld, bitrate=%ld\n", 
            params.maxWidth, params.maxHeight, params.maxBitRate);

    GST_LOG("opening video encoder \"%s\"\n", videnc1->codecName);
    videnc1->hVe1 = Venc1_create(videnc1->hEngine, (Char*)videnc1->codecName,
                      &params, &dynParams);

    if (videnc1->hVe1 == NULL) {
        GST_ELEMENT_ERROR(videnc1, STREAM, CODEC_NOT_FOUND,
        ("failed to create video encoder: %s\n", videnc1->codecName), (NULL));
        GST_LOG("closing codec engine\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Create a circular input buffer */
    if (videnc1->numInputBufs == 0) {
        videnc1->numInputBufs = 2;
    }

    /* Check if we need to create circular buffer with different size */
    if (videnc1->upstreamBufSize > 0) {
        inBufSize = videnc1->upstreamBufSize;
    }
    else {
        inBufSize = Venc1_getInBufSize(videnc1->hVe1);
    }

    videnc1->circBuf = gst_ticircbuffer_new(inBufSize, 
                            videnc1->numInputBufs, TRUE);

    if (videnc1->circBuf == NULL) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, NO_SPACE_LEFT,
        ("failed to create circular input buffer\n"), (NULL));
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Create codec output buffers */
    GST_LOG("creating output buffer table\n");
    gfxAttrs.colorSpace     = videnc1->colorSpace;
    gfxAttrs.dim.width      = videnc1->width;
    gfxAttrs.dim.height     = videnc1->height;
    gfxAttrs.dim.lineLength = BufferGfx_calcLineLength(
                                  gfxAttrs.dim.width, gfxAttrs.colorSpace);

    gfxAttrs.bAttrs.memParams.align = 128;

    /* By default, new buffers are marked as in-use by the codec */
    gfxAttrs.bAttrs.useMask = gst_tidmaibuffer_CODEC_FREE;

    if (videnc1->numOutputBufs == 0) {
        videnc1->numOutputBufs = 3;
    }

    videnc1->hOutBufTab = gst_tidmaibuftab_new(videnc1->numOutputBufs,
        Venc1_getOutBufSize(videnc1->hVe1),
        BufferGfx_getBufferAttrs(&gfxAttrs));

    if (videnc1->hOutBufTab == NULL) {
        GST_ERROR("failed to create output buffers\n");
        gst_tividenc1_exit_video(videnc1);
        return FALSE;
    }

    /* Display buffer contents if displayBuffer=TRUE was specified */
    gst_ticircbuffer_set_display(videnc1->circBuf, videnc1->displayBuffer);

    return TRUE;
}

/******************************************************************************
 * gst_tividenc1_populate_codec_header
 *  This function populates codec_data field for H.264.
 *****************************************************************************/
void gst_tividenc1_populate_codec_header (GstTIVidenc1 *videnc1, 
    Buffer_Handle hDstBuf)
{
    if (gst_is_h264_encoder(videnc1->codecName)) {
        if (!videnc1->byteStream) {
            /* Generate codec_data field for packetized stream */
            if (!videnc1->codec_data) {
                videnc1->codec_data = gst_h264_create_codec_data(hDstBuf);
            }
        }
    }
}

/******************************************************************************
 * gst_tividenc1_parse_and_queue 
 *  Function parses and push the output buffer .
 *****************************************************************************/
GstFlowReturn  gst_tividenc1_parse_and_push (GstTIVidenc1 *videnc1, 
    GstBuffer *outBuf)
{
    guint8 *data;
    guint   len;

    /* perform H.264 specific parsing before pushing the data */
    if (gst_is_h264_encoder(videnc1->codecName)) {

        /* convert byte-stream to packetized */
        if ((!videnc1->byteStream) && (videnc1->codec_data)) {

            /* Prefix the NALU with length field */
            data = GST_BUFFER_DATA(outBuf);
            len = GST_BUFFER_SIZE(outBuf)-4;
           
            data[3] = len & 0xff;
            data[2] = (len >> 8) & 0xff;
            data[1] = (len >> 16) & 0xff;
            data[0] = (len >> 24) & 0xff;
        }
    }

    return gst_pad_push(videnc1->srcpad, outBuf);
}

/******************************************************************************
 * gst_tividenc1_encode_thread
 *     Call the video codec to process a full input buffer
 ******************************************************************************/
static void* gst_tividenc1_encode_thread(void *arg)
{
    GstTIVidenc1        *videnc1       = GST_TIVIDENC1(gst_object_ref(arg));
    GstBuffer           *encDataWindow  = NULL;
    BufferGfx_Attrs     gfxAttrs        = BufferGfx_Attrs_DEFAULT;
    void                *threadRet      = GstTIThreadSuccess;
    Buffer_Handle       hDstBuf, hInBuf = NULL;
    Int32               encDataConsumed;
    GstClockTime        encDataTime;
    GstClockTime        frameDuration;
    Buffer_Handle       hEncDataWindow;
    GstBuffer           *outBuf;
    Int                 bufIdx;
    Int                 ret;

    /* Calculate the duration of a single frame in this stream */
    frameDuration = gst_tividenc1_frame_duration(videnc1);

    /* Initialize codec engine */
    ret = gst_tividenc1_codec_start(videnc1);

    /* Notify main thread that is ok to continue initialization */
    Rendezvous_meet(videnc1->waitOnEncodeThread);
    Rendezvous_reset(videnc1->waitOnEncodeThread);

    if (ret == FALSE) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, FAILED, 
        ("failed to start codec\n"), (NULL));
        goto thread_exit;
    }

    /* set graphics attrs for input buffer */
    gfxAttrs.dim.width  = videnc1->width;
    gfxAttrs.bAttrs.reference = TRUE;
    gfxAttrs.dim.height  = videnc1->height;
    gfxAttrs.colorSpace  = videnc1->colorSpace;
    gfxAttrs.dim.lineLength = BufferGfx_calcLineLength(videnc1->width,
                                    videnc1->colorSpace);

    /* DM6467: set the colorspace to YUV420PSEMI, this is mainly because on 
     * dm6467 circular buffer will perform the color conversion from 422-420.
     */
    if ((videnc1->device == Cpu_Device_DM6467) && 
            videnc1->colorSpace == ColorSpace_YUV422PSEMI) {
        gfxAttrs.colorSpace  = ColorSpace_YUV420PSEMI;
    }
 
    /* allocate reference input buffer */
    hInBuf = Buffer_create(1, 
                BufferGfx_getBufferAttrs(&gfxAttrs));
    if (hInBuf == NULL) {
        GST_ELEMENT_ERROR(videnc1, RESOURCE, NO_SPACE_LEFT,
        ("failed to allocate ccv buffer\n"), (NULL));
        goto thread_failure;
    }

    /* Main thread loop */
    while (TRUE) {

        /* Obtain an encoded data frame */
        encDataWindow  = gst_ticircbuffer_get_data(videnc1->circBuf);
        encDataTime    = GST_BUFFER_TIMESTAMP(encDataWindow);
        hEncDataWindow = GST_TIDMAIBUFFERTRANSPORT_DMAIBUF(encDataWindow);

        /* If we received a data frame of zero size, there is no more data to
         * process -- exit the thread.  If we weren't told that we are
         * draining the pipeline, something is not right, so exit with an
         * error.
         */
        if (GST_BUFFER_SIZE(encDataWindow) == 0) {
            GST_LOG("no video data remains\n");
            if (!videnc1->drainingEOS) {
                goto thread_failure;
            }
            goto thread_exit;
        }

        /* Obtain a free output buffer for the encoded data */
        if (!(hDstBuf = gst_tidmaibuftab_get_buf(videnc1->hOutBufTab))) {
            GST_ELEMENT_ERROR(videnc1, RESOURCE, READ,
                ("failed to get a free contiguous buffer from BufTab\n"),
                (NULL));
            goto thread_failure;
        }

        /* Make sure the whole buffer is used for output */
        BufferGfx_resetDimensions(hDstBuf);

        /* set the input buffer size */ 
        if (videnc1->device == Cpu_Device_DM6467 &&
             videnc1->colorSpace == ColorSpace_YUV422PSEMI) {
            Buffer_setSize(hInBuf, videnc1->width * videnc1->height * 3 / 2);
        }
        else {
            Buffer_setSize(hInBuf, GST_BUFFER_SIZE(encDataWindow));
        }

        /* Update the user pointer for input buffer. */        
        Buffer_setNumBytesUsed(hInBuf, Buffer_getSize(hInBuf));
        Buffer_setUserPtr(hInBuf, Buffer_getUserPtr(hEncDataWindow));

        /* Invoke the video encoder */
        GST_LOG("invoking the video encoder\n");
        ret   = Venc1_process(videnc1->hVe1, hInBuf, hDstBuf);

        encDataConsumed = GST_BUFFER_SIZE(encDataWindow);

        if (ret < 0) {
            GST_ELEMENT_ERROR(videnc1, STREAM, ENCODE,
            ("failed to encode video buffer\n"), (NULL));
            goto thread_failure;
        }

        /* If no encoded data was used we cannot find the next frame */
        if (ret == Dmai_EBITERROR && encDataConsumed == 0) {
            GST_ELEMENT_ERROR(videnc1, STREAM, ENCODE,
            ("fatal bit error\n"), (NULL));
            goto thread_failure;
        }

        if (ret > 0) {
            GST_LOG("Venc1_process returned success code %d\n", ret); 
        }

        /* Release the reference buffer, and tell the circular buffer how much
         * data was consumed.
         */
        ret = gst_ticircbuffer_data_consumed(videnc1->circBuf, encDataWindow,
                  encDataConsumed);
        encDataWindow = NULL;

        if (!ret) {
            goto thread_failure;
        }

        /* Populate codec header */
        gst_tividenc1_populate_codec_header(videnc1, hDstBuf);

        /* Set the source pad capabilities based on the encoded frame
         * properties.
         */
        gst_tividenc1_set_source_caps(videnc1, hDstBuf);

        /* Create a DMAI transport buffer object to carry a DMAI buffer to
         * the source pad.  The transport buffer knows how to release the
         * buffer for re-use in this element when the source pad calls
         * gst_buffer_unref().
         */
        outBuf = gst_tidmaibuffertransport_new(hDstBuf, videnc1->hOutBufTab);
        gst_buffer_set_data(outBuf, GST_BUFFER_DATA(outBuf),
            Buffer_getNumBytesUsed(hDstBuf));
        gst_buffer_set_caps(outBuf, GST_PAD_CAPS(videnc1->srcpad));

        /* If we have a valid time stamp, set it on the buffer */
        if (videnc1->genTimeStamps &&
            GST_CLOCK_TIME_IS_VALID(encDataTime)) {
            GST_LOG("video timestamp value: %llu\n", encDataTime);
            GST_BUFFER_TIMESTAMP(outBuf) = encDataTime;
            GST_BUFFER_DURATION(outBuf)  = frameDuration;
        }
        else {
            GST_BUFFER_TIMESTAMP(outBuf) = GST_CLOCK_TIME_NONE;
        }

        /* Tell circular buffer how much time we consumed */
        gst_ticircbuffer_time_consumed(videnc1->circBuf, frameDuration);

        /* Parse and Push the transport buffer to the source pad */
        GST_LOG("pushing display buffer to source pad\n");
        if (gst_tividenc1_parse_and_push(videnc1, outBuf) != GST_FLOW_OK) {
            GST_DEBUG("push to source pad failed\n");
            goto thread_failure;
        }

        /* Release buffers no longer in use by the codec */
        Buffer_freeUseMask(hDstBuf, gst_tidmaibuffer_CODEC_FREE);
    }

thread_failure:

    gst_tithread_set_status(videnc1, TIThread_CODEC_ABORTED);
    gst_ticircbuffer_consumer_aborted(videnc1->circBuf);
    threadRet = GstTIThreadFailure;

thread_exit: 

    /* Re-claim any buffers owned by the codec */
    bufIdx = BufTab_getNumBufs(GST_TIDMAIBUFTAB_BUFTAB(videnc1->hOutBufTab));

    while (bufIdx-- > 0) {
        Buffer_Handle hBuf = BufTab_getBuf(
            GST_TIDMAIBUFTAB_BUFTAB(videnc1->hOutBufTab), bufIdx);
        Buffer_freeUseMask(hBuf, gst_tidmaibuffer_CODEC_FREE);
    }

    /* Release the last buffer we retrieved from the circular buffer */
    if (encDataWindow) {
        gst_ticircbuffer_data_consumed(videnc1->circBuf, encDataWindow, 0);
    }

    /* We have to wait to shut down this thread until we can guarantee that
     * no more input buffers will be queued into the circular buffer
     * (we're about to delete it).  
     */
    Rendezvous_meet(videnc1->waitOnEncodeThread);
    Rendezvous_reset(videnc1->waitOnEncodeThread);

    /* Notify main thread that we are done draining before we shutdown the
     * codec, or we will hang.  We proceed in this order so the EOS event gets
     * propagated downstream before we attempt to shut down the codec.  The
     * codec-shutdown process will block until all BufTab buffers have been
     * released, and downstream-elements may hang on to buffers until
     * they get the EOS.
     */
    Rendezvous_force(videnc1->waitOnEncodeDrain);

    /* Stop codec engine */
    if (gst_tividenc1_codec_stop(videnc1) < 0) {
        GST_ERROR("failed to stop codec\n");
    }

    gst_object_unref(videnc1);

    GST_LOG("exit video encode_thread (%d)\n", (int)threadRet);
    return threadRet;
}

/******************************************************************************
 * gst_tividenc1_drain_pipeline
 *    Wait for the encode thread to finish processing queued input data.
 ******************************************************************************/
static void gst_tividenc1_drain_pipeline(GstTIVidenc1 *videnc1)
{
    gboolean checkResult;

    /* If the encode thread hasn't been created, there is nothing to drain. */
    if (!gst_tithread_check_status(
             videnc1, TIThread_CODEC_CREATED, checkResult)) {
        return;
    }

    videnc1->drainingEOS = TRUE;
    gst_ticircbuffer_drain(videnc1->circBuf, TRUE);

    /* Tell the encode thread that it is ok to shut down */
    Rendezvous_force(videnc1->waitOnEncodeThread);

    /* Wait for the encoder to finish draining */
    Rendezvous_meet(videnc1->waitOnEncodeDrain);
}


/******************************************************************************
 * gst_tividenc1_frame_duration
 *    Return the duration of a single frame in nanoseconds.
 ******************************************************************************/
static GstClockTime gst_tividenc1_frame_duration(GstTIVidenc1 *videnc1)
{
    /* Default to 29.97 if the frame rate was not specified */
    if (videnc1->framerateNum == 0 && videnc1->framerateDen == 0) {
        GST_WARNING("framerate not specified; using 29.97fps");
        videnc1->framerateNum = 30000;
        videnc1->framerateDen = 1001;
    }

    return (GstClockTime)
        ((1 / ((gdouble)videnc1->framerateNum/(gdouble)videnc1->framerateDen))
         * GST_SECOND);
}


/******************************************************************************
 * Custom ViM Settings for editing this file
 ******************************************************************************/
#if 0
 Tabs (use 4 spaces for indentation)
 vim:set tabstop=4:      /* Use 4 spaces for tabs          */
 vim:set shiftwidth=4:   /* Use 4 spaces for >> operations */
 vim:set expandtab:      /* Expand tabs into white spaces  */
#endif
